/*
 * File: epimsg.c
 * Author: Jim MacNair
 * Version: 1.1
 * 
 * V1.1
 * 
 *  1) fixed bug with moving message to allocated storage rather than passed storage
 *
 */

#define CICS_W32
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cics_epi.h>
#include "epimsg.h"

/*******************************************************************/
/*                                                                 */
/* This routine will take an EPI return code and store a           */
/*  descriptive error message in the passed message area.          */
/*  The message area should be allocated as:                       */
/*                                                                 */
/*  char errmsg[CICS_3270_MAX_EPI_MSG_LEN];                        */
/*                                                                 */
/*  in the calling program.                                        */
/*                                                                 */
/*******************************************************************/

void CICS_3270_GetEpiErrMsg(int rc,
                            cics_ushort_t TermIndex,
                            char *msg)

{
   CICS_EpiSysError_t CICSPTR   *err;
   int                          rcode;

   switch(rc)
   {
       case CICS_EPI_NORMAL:
              strcpy(msg,"Normal Return");
              break;

       /* didn't call EpiInitialize                             */
       /* Initialize sets up client data structures, and starts */
       /* a thread to handle replies from the client.  It does  */
       /* not connect to any servers.                           */

       case CICS_EPI_ERR_NOT_INIT:
              strcpy(msg,"EPI not initialized - call EpiInitialize");
              break;

       case CICS_EPI_ERR_BAD_INDEX:
              strcpy(msg,"Invalid Terminal Index");
              break;

       /* General error code for miscellaneous */
       /* errors.                              */
       /* Try and get a secondary error code,  */
       /* although the secondary codes are not */
       /* documented                           */

       case CICS_EPI_ERR_FAILED:
          err = malloc(sizeof(CICS_EpiSysError_t));
          rcode = CICS_EpiGetSysError(TermIndex, err);

          if (rc == CICS_EPI_NORMAL)
          {
             sprintf(msg,"EPI Call Failed, cause %d value %d %s",
                     err->Cause,err->Value,err->Msg);
          }
          else
          {
             strcpy(msg,"EPI Call Failed");
          } /* endif */

          free(err);
          break;

       /* Bad version number in first parameter in EpiInitialize */

       case CICS_EPI_ERR_VERSION:
              strcpy(msg,"Invalid Version - bad first parm in EpiInitialize call");
              break;

       /* Second call to EpiInitialize - it should only be */
       /* called once.  When the program is finished, it   */
       /* should call delete all terminals and then        */
       /* call EpiTerminate                                */

       case CICS_EPI_ERR_IS_INIT:
              strcpy(msg,"EPI is already initialized - second call to EpiInitialize");
              break;

       /* No systems in INI file */

       case CICS_EPI_ERR_NO_SYSTEMS:
              strcpy(msg,"No systems defined - check CICSCLI.INI");
              break;

       /* The data structure passed could not hold all */
       /* the systems that are defined to this  client */
       /* Returned from EpiListSystems                 */

       case CICS_EPI_ERR_MORE_SYSTEMS:
              strcpy(msg,"More systems - passed data structure cold not hold all systems");
              break;

       /* Default limits for DOS and Win 3.1 are 15 */
       /* AIX client has limit of about 100-200+ */

       case CICS_EPI_ERR_MAX_TERMS:
              strcpy(msg,"No more terminals available");
              break;

       /* On EpiAddTerminal, the system is not defined */
       /* in the CICSCLI.INI file.                     */

       case CICS_EPI_ERR_SYSTEM:
              strcpy(msg,"System not defined in CICSCLI.INI");
              break;

       /* Issued DelTerminal but transaction is running */

       case CICS_EPI_ERR_TRAN_ACTIVE:
              strcpy(msg,"Transaction active - cannot issue DelTerminal");
              break;

       /* Start Tran and Reply require at least an aid */
       /* byte and usually two cursor location bytes   */

       case CICS_EPI_ERR_NO_DATA:
              strcpy(msg,"Not enough data for EpiStartTran or EpiReply");
              break;

       /* StartTran issued when transaction already active  */
       /* The transaction was started with a start tran     */
       /* This is probably a logic error in the application */

       case CICS_EPI_ERR_TTI_ACTIVE:
              strcpy(msg,"Transaction already started with EpiStartTran");
              break;

       /* StartTran issued when transaction already active  */
       /* The transaction was started by an ATI request     */
       /* by the host.  This could be a logic error or      */
       /* a race condition, since the application may not   */
       /* have seen the START_ATI event when it issued      */
       /* the EpiStartTran.                                 */

       case CICS_EPI_ERR_ATI_ACTIVE:
              strcpy(msg,"ATI transaction active - cannot issue EpiStartTran");
              break;

       /* no converse event was received */

       case CICS_EPI_ERR_NO_CONVERSE:
              strcpy(msg,"EpiReply issued, but no outstanding converse");
              break;

       case CICS_EPI_ERR_ATI_STATE:
              strcpy(msg,"ATI State Error - Bad value for second parameter of EpiATIState");
              break;

       /* GetEvent was issued with no wait option, and no */
       /* events are pending - this is a normal return    */
       /* code and does not indicate an error             */

       case CICS_EPI_ERR_NO_EVENT:
              strcpy(msg,"No events are pending");
              break;

       /* There are more events waiting to be processed   */
       /* after this one.  Another GetEvent should be     */
       /* issued after processing of the current one is   */
       /* complete.  This return code is commonly seen    */
       /* with a send followed by an end tran event.      */

       case CICS_EPI_ERR_MORE_EVENTS:
              strcpy(msg,"More events are pending after this");
              break;

       case CICS_EPI_ERR_MORE_DATA:
              strcpy(msg,"More data is available");
              break;

       /* The wait parameter is an enum value, and can    */
       /* have bad values, even though it can only be     */
       /* specified as true or false.                     */

       case CICS_EPI_ERR_WAIT:
              strcpy(msg,"Wait parameter on EpiGetEvent is invalid value");
              break;

       /* Connection attempt failed.  The system is */
       /* properly defined in the CICSCLI.INI file. */

       case CICS_EPI_ERR_SERVER_DOWN:
              strcpy(msg,"The CICS Server is down");
              break;

       case CICS_EPI_ERR_SENSE_CODE:
              strcpy(msg,"Sense code is available");
              break;

       /* CTIN needs a user id and password to run, but */
       /* no userid and password was specified on the   */
       /* command line or in the security popup window. */
       /* The security popup may be suppressed by the   */
       /* /Q switch on the CICSCLI command, if the      */
       /* client is started from the command line.      */

       case CICS_EPI_ERR_SECURITY:
              strcpy(msg,"Security error trying to run the host CTIN transaction to install the terminal");
              break;

       /* Issued EPI call while in callback routine     */
       /* This is an application error.                 */

       case CICS_EPI_ERR_IN_CALLBACK:
              strcpy(msg,"Error in callback routine - tried to issue EPI call");
              break;

       case CICS_EPI_ERR_NULL_PARM:
              strcpy(msg,"Parameter cannot be NULL");
              break;

       default:
              sprintf(msg,"Unknown return code - %d",rc);
              break;

   } /* endswitch */
}

/*-------------*/
/* End Of File */
/*-------------*/
