/*
 * Author: Jim MacNair
 * Version: 1.1
 * Date: 04/15/98
 *
 * Description:
 * This program is a special version of the 3270 simulator program.
 * It is designed to accept 3270 input from files rather than 
 * from a live 3270 system.  The program takes a command file
 * as input.  All 3270 data streams must be read in from files
 * using the READRAW command.  The 3270 data buffer can be 
 * modified but no input can be sent.
 *
 * The program uses an input file.  The format of the command is:
 *
 *    file3270 input-file <cics-region|* <netname|*> <term model> >
 *
 * The input file contains command statements.  Each command
 * statement must be on a separate line.  The valid commands are:
 *
 *   MODFIELD row column data
 *   MODNUMFLD field-number data
 *   KEYIN [keystroke-data]
 *   SETCURSOR row column
 *   CLEAR
 *   PRINTFLD row column <filename>
 *   PRINTNUMFLD field-number data
 *   DUMPSCR  <filename>
 *   DUMPFLDS <filename>
 *   DUMPRAW  filename
 *   READRAW  filename
 *   WRITEFIELDS <filename>
 *   WRITEPW  filename <prefix>
 *   TRACE OFF | ON <filename>
 *
 * MODFIELD is used to alter a field in the screen image.
 *
 * CLEAR is a shorthand for KEYIN with predefined aid.  The 
 * screen buffer will be cleared.
 *
 * The DUMPSCR and DUMPFLDS are used to write the current contents
 * of the screen to a file.  If no file is specified, then the
 * output is sent to stdout.
 *
 * The input file will be processed.  It must include READRAW
 * statements to load the 3270 screen buffer.  Any MODFIELD,
 * DUMPSCR, DUMPFLDS or PRINTFLD statements encountered will
 * be processed.
 *
 * The PRINTFLD, DUMPSCR and DUMPFLDS write out information
 * about the current 3270 screen image.  PRINTFLD will write
 * the current begining location (row and column), attribute,
 * data length and actual data for the field which contains
 * the given row and column to stdout.  DUMPSCR will write
 * the current screen image to a file, substituting an @
 * character for each attribute byte.  DUMPFLDS will write
 * one line with the location and characteristics of each
 * attribute byte to a file.
 *
 * The WRITEFIELDS statement will write a description of
 * each field on a separate line.  The line contains the
 * location (row and column), length and up to 30 characters 
 * of data for each field, plus the basic field attributes.
 *
 * The WRITEPW statement will update an Interspace (TM)
 * catalog file with information about the current screen.
 * Only unprotected fields are added to the catalog.  If
 * an unprotected field is preceded on the same line by
 * a protected field, then this is likely to be a label
 * and an attempt is made to use this label as a field
 * name.  Otherwise, a variable name of the format 
 * FIELDnnn is assigned.  This command will update the
 * catalog if it exists, and will create a new catalog
 * if the cataog does not exist.  It will rename the 
 * existing catalog to a file extension of "bak", and
 * the updated catalog will retain the extension of
 * "cat".  
 *
 * The READRAW command will read a file containing a valid   
 * 3270 datastream and load it into the screen buffer.
 * The file must contain a valid 3270 write command and a
 * valid write control character (WCC) in the first two
 * positions, followed by the orders and data.  This 
 * command is aimed at customers who want to feed a 3270 datastream
 * from another program.  
 *
 * Lines beginning with a "*", "'", "/" or ";" will be
 * treated as comments and ignored.
 *
 * Rrace information will be written if the
 * DEBUG environment variable is set.  If used, the DEBUG
 * variable should be set to the name of the desired DEBUG
 * file.  To direct trace information to stdout, set
 * DEBUG=stdout.
 *
 * Command lines can be a maximum of 256 characters long.  Lines longer
 * than this will be truncated.
 *
 * This program is written for the Windows/NT environment, but should
 * be portable to other platforms with little or no change.
 *
 */

/*
 *
 * Initial release: V1.1
 *
 * Updates: none
 *
 */

#define CICS_W32
#define u_short_t cics_ushort_t
#define u_byte_t  cics_ubyte_t
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "c3270.h"
#include "errorlog.h"

static char Version[]=
"@(#)SIM3270 V1.1 - 3270 file Simulator - Jim MacNair ("__DATE__")";

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#define MAX_LINE   255          /* Maximum length of command */
#define ATTR_CHAR  '@'          /* default screen dump attribute char */
#define NULL_CHAR  '_'          /* default screen dump null char */
#define SCREEN_FILE_NAME   "scrfile.txt\0"
#define NEW_TRAN_DATA "\x027  \0"
#define PWCAT_EXT ".cat"
#define TEMP_EXT  ".tmp"
#define BAK_EXT   ".bak"

/*
 * Line number within the command File
 */

   int          lineNum=0;

/*
 * Name of Screen Dump File
 */

   char         ATIFileName[255]="";
   int          ATIEnabled=0;

/*
 * Name of Default Screen Dump File
 */

   char         scrnFile[255]="\0";;

/*
 *
 * Routine to check if a file name is specified for a
 * screen dump file.  If it is, then that name is used,
 * otherwise the default name defined above is used.
 *
 */
void buildScrFileName(char *cmdData)

{
   if ((cmdData == NULL) || ((*cmdData) == '\0'))
   {
      strcpy(scrnFile,SCREEN_FILE_NAME);
   }
   else
   {
      strcpy(scrnFile,cmdData);
   } /* endif */
}

/*
 *
 * Routine to find the next variable in the command line
 * and terminate, moving the data pointer.
 *
 */

void parseVar(char *cmdData, int *nextPos)
{
   char         *newData;

   /*
    * Skip over any leading blanks
    */

   while (cmdData[0] == ' ')
   {
      cmdData++;
   } /* endwhile */

   if ((cmdData != NULL) && ((*cmdData) != '\0'))
   {
      newData = strchr(cmdData, ' ');  /* find trailing blank */
      if (newData != NULL)
      {
         newData[0] = '\0';  /* Terminate the string */
      } /* endif */

      newData++;             /* move the pointer */
      (*nextPos) = (int) (newData - cmdData);
   } /* endif */

   return;
}

/*
 *
 * Routine to find the next number in the command line
 * and return it as an integer, moving the data pointer.
 *
 */

int parseNumber(char *cmdData, int *nextPos)
{
   int          i=0;
   char         *newData;

   /* initialize next position offset to 0, in case */
   /* there is no parameter to process              */
   *nextPos = 0;

   /*
    * Skip over any leading blanks
    */

   while (cmdData[0] == ' ')
   {
      cmdData++;
   } /* endwhile */

   /*
    * Find the end of the number
    */

   if ((cmdData != NULL) && ((*cmdData) != '\0'))
   {
      newData = strchr(cmdData, ' ');  /* find trailing blank */
      if (newData != NULL)
      {
         newData[0] = '\0';  /* Terminate the string */
      }
      else
      {
         newData = cmdData + strlen(cmdData) - 1;
      } /* endif */

      if ((cmdData[0] >= '0') && (cmdData[0] <= '9'))
      {
         i = atoi(cmdData);     /* get the value */
      } /* endif */

      newData++;             /* move the pointer */
      (*nextPos) = (int) (newData - cmdData);
   } /* endif */

   return i;
}

/*
 *
 * Routine to handle TRACE commands.
 *
 */

void processTrace(char *cmdData)
{
   int                 rc;
   char               *fileName;
   char               *endfileName;

   /* Skip leading blanks in cmdData */
   while (cmdData[0] == ' ')
   {
           cmdData++;
   }

   /* Find the end of the first parameter */
   fileName = strchr(cmdData,' ');
   if (fileName != NULL)
   {
      fileName[0] = '\0';   /* terminate the first parm */
      fileName++;           /* move past string terminator */

      while (fileName[0] == ' ')
      {
         fileName++;  /* skip leading blanks */
      }

      /* look for trailing  blanks */
      endfileName = strchr(fileName, ' ');
      if (endfileName != NULL)
      {
         endfileName[0] = '\0';
      } /* endif */
   }

   /* Now set a variable to indicate if we are */
   /* turning trace on or off or querying status */

   if (strcmp(cmdData,"ON\0") == 0)
   {
      rc = CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_ON, fileName);
   }
   else
   {
      if (strcmp(cmdData,"APPEND\0") == 0)
      {
         rc = CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_APPEND, fileName);
      }
      else
      {
         if (strcmp(cmdData,"OFF\0") == 0)
         {
            rc = CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_OFF, NULL);
         }
         else
         {
            rc = CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_INQ, NULL);
         }
      }
   }

 if (rc)
 {
    printf("Trace is enabled\n");
 }
 else
 {
    printf("Trace is off\n");
 }
}

/*
 *
 * Routine to handle SETCURSOR commands.
 *
 */

void processSetCursor(char *cmdData)

{
   int                 row=0;
   int                 col=0;
   int                 nextPos;

   /* Get the row and column from the command data */
   row = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;

   /* skip over any blanks */
   while (*cmdData == ' ')
   {
      cmdData++;
   }

   if (strlen(cmdData) > 0)
   {
      col = parseNumber(cmdData,&nextPos);

      /* Print the characteristics of the field */
      /* containing the desired location */
      CICS_3270_SetCursor(row, col);
   }
   else
   {
      CICS_3270_Write_Log_Num(lineNum,"missing or invalid parameters for SETCURSOR");
   }
}

/*
 *
 * Routine to handle READRAW commands.
 *
 *
 * A file containing a complete 3270 data stream is read from a
 * file and then applied to the input buffer.  The file must begin
 * with a valid write command followed by a write control character.
 * The file cannot exceed 32767 characters.
 *
 */

void readRawFile(char *cmdData)

{
	FILE *inpfile;
	char errline[255];
	char databuffer[32768];
	int  len;

	/* open the file for binary input */
	inpfile = fopen(cmdData, "rb");
	if (inpfile == NULL)
	{
		/* write an error message to the log and exit */
		sprintf(errline, "unable to open 3270 input file - %s", cmdData);
		CICS_3270_Write_Log_Num(lineNum,errline);
		return;
	}

	/* read the file into the buffer */
	len = fread(databuffer, sizeof(char), sizeof(databuffer), inpfile);

	CICS_3270_Parse(databuffer, len);

	/* close the raw input file */
	fclose(inpfile);
}

/*
 *
 * Routine to handle WRITEFIELDS commands.
 *
 */

void processWriteFields(char *cmdData)

{
	/* write a file containing a line per field */
	CICS_3270_WriteFields(cmdData);
}

/*
 *
 * Routine to create an empty PlanetWorks catalog file.
 *
 */

int createPWcat(char *fileName)

{
	FILE    *catfile;

	/* create the catalog file */
	catfile = fopen(fileName, "w");
	if (catfile == NULL)
	{
		/* unable to create file */
		return 100;
	}

	/* Write some comments to identify the file */
	fprintf(catfile, "#\n");
	fprintf(catfile, "# Interspace service definition file for CICS 3270\n");
	fprintf(catfile, "# Captured by Sim3270 program from IBM SupportPac CA04\n");
	fprintf(catfile, "#\n");
	fprintf(catfile, "# Fields are defined here\n");
	fprintf(catfile, "# NAME		TYPE	MEMORY	DISPLAY	DESCRIPTION\n");

	/* Indicate the beginning of the field definitions */
	fprintf(catfile, "*FIELDS\n");

	/* insert a service definition for the transaction */
	fprintf(catfile, "#\n");
	fprintf(catfile, "# Service Definitions\n");
	fprintf(catfile, "#\n");
	fprintf(catfile, "*SERVICES\n");
	fprintf(catfile, "#\n");

	/* Close the file */
	fclose(catfile);

	/* indicate normal return */
	return 0;
}

/*
 *
 * Routine to handle WRITEPW (write interspace catalog files) commands.
 *
 */

int processWritePW(char *cmdData, char *trnid, char *syst)

{
	FILE *oldcat;
	FILE *newcat;
	char filename[255];
	char newfilename[255];
	char bakfilename[255];
	char pref[255];
	char errline[255];
	char *prefix;
	char *fileext;
	int  foundbak = 0;
	int  rc;

	/* get the catalog file name */
	strcpy(filename, cmdData);

	/* look for a blank following the file name */
	prefix = strchr(filename, ' ');

	/* did we find a blank? */
	if (prefix == NULL)
	{
		/* point to a null prefix if none is specified */
		prefix = "";
	}
	else
	{
		/* terminate the file name string */
		prefix[0] = '\0';

		/* skip the null character */
		prefix++;

		/* skip any leading blanks */
		while (prefix[0] == ' ')
		{
			prefix++;
		}
	}

	/* save the prefix */
	strcpy(pref, prefix);

	/* Parse the catalog file name, and remove the extension, if any */
	fileext = strchr(filename, '.');

	/* did we find one? */
	if (fileext != NULL)
	{
		/* remove the file extension */
		fileext[0] = '\0';
	}

	/* get the new catalog name */
	strcpy(newfilename, filename);
	strcpy(bakfilename, filename);

	/* append the right file extensions for the old and new catalogs */
	strcat(filename, PWCAT_EXT);
	strcat(newfilename, TEMP_EXT);
	strcat(bakfilename, BAK_EXT);

	/* try to open the catalog, to see if it already exists */
	oldcat = fopen(filename, "r");
	if (oldcat == NULL)
	{
		/* Check for a .bak file */
		oldcat = fopen(bakfilename, "r");
		if (oldcat == NULL)
		{
			/* couldn't open the file, let's try to create one */
			if (createPWcat(filename) == 0)
			{
				/* now, try to open it again */
				oldcat = fopen(filename, "r");
				if (oldcat == NULL)
				{
					/* unable to open the catalog file */
					return 100;
				}
			}
			else
			{
				/* unable to create catalog file */
				return 100;
			}
		}
		else
		{
			/* found a file with a .bak extension */
			foundbak = 1;
		}
	}

	/* try to open the temporary new catalog */
	newcat = fopen(newfilename, "w");
	if (newcat == NULL)
	{
		/* close the old catalog */
		fclose(oldcat);

		/* problem opening the temporary new catalog file */
		return 100;
	}

	/* write a Planetworks Interspace(R) catalog file */
	CICS_3270_WritePW(oldcat, newcat, pref, trnid, syst);

	/* close the two files */
	fclose(oldcat);
	fclose(newcat);

	/* Lastly, rename the files, leaving the old catalog */
	/* with an extension of .bak                         */
	/* Do this in several steps, first erasing any old   */
	/* .bak files, then renaming the current file to     */
	/* .bak, and finally renaming the temporary file     */
	/* to an .cat extension.                             */

	/* check if we couldn't find the .cat file and used the .bak file */
	if (foundbak == 0)
	{
		/* delete the old .bak file */
		rc = remove(bakfilename);
		if ((rc == 0) || (rc = -1))
		{
			/* delete worked or no such file, now try to rename the .cat file to .bak */
			rc = rename(filename, bakfilename);
			if (rc != 0)
			{
				/* can't rename the original file - return error */
				sprintf(errline, "unable to rename %s to %s - %d\n", filename, bakfilename, rc);

				/* write message to error log */
				CICS_3270_Write_Log_Num(lineNum, errline);

				return 100;
			}
		}
		else
		{
			/* can't erase the old .bak file - return error */
			sprintf(errline, "unable to delete the file %s - %d\n", bakfilename, rc);

			/* write message to error log */
			CICS_3270_Write_Log_Num(lineNum, errline);

			return 100;
		}
	}

	/* rename the .tmp file to .cat to finish the operation */
	rc = rename(newfilename, filename);

	/* check if the rename worked */
	if (rc != 0)
	{
		/* can't rename the new file to the correct extension */
		sprintf(errline, "unable to rename %s to %s - %d\n", newfilename, filename, rc);

		/* write message to error log */
		CICS_3270_Write_Log_Num(lineNum, errline);

		return 100;
	}

	return 0;
}

/*
 *
 * Routine to handle PRINTNUMFLD commands.
 *
 */

void processPrintNumField(char *cmdData)

{
   int     row=0;
   int     col=0;
   int     num=0;
   int     nextPos;

   /* Get the field number from the command data */
   num = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;
   parseVar(cmdData, &nextPos);

   /* get the row and column of the field */
   CICS_3270_GetFieldLocByNum(num, &row, &col);

   if (row != -1)
   {
      /* Print the characteristics of the field */
      /* containing the desired location */
      CICS_3270_PrintField(row, col, cmdData);
   }
   else
   {
      CICS_3270_Write_Log_Num(lineNum, "Invalid print field number");
   }
}

/*
 *
 * Routine to handle PRINTFIELD commands.
 *
 */

void processPrintField(char *cmdData)
{
   int                 row=0;
   int                 col=0;
   int                 nextPos;

   /* Get the row and column from the command data */
   row = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;
   col = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;
   parseVar(cmdData, &nextPos);

   /* Print the characteristics of the field */
   /* containing the desired location */
   CICS_3270_PrintField(row, col, cmdData);
}

/*
 *
 * Routine to handle KEYIN commands.
 *
 */

void processKeyin(char *cmdData)
{
   /* Process the keyin request */
   CICS_3270_Keyin(cmdData);
}

/*
 *
 * Routine to handle MODFIELD commands.
 *
 */

void processModField(char *cmdData)
{
   int          row=0;
   int          col=0;
   int          lngth=0;
   int          nextPos;

   /* Get the row and column from the command data */
   row = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;
   col = parseNumber(cmdData,&nextPos);
   /* Point to the new data */
   cmdData += nextPos;

   /* Get the length of the new data */
   lngth = strlen(cmdData);

   /* Modify the specified field, using C3270.C */
   CICS_3270_ModifyField(row,col,cmdData,lngth);
}

/*
 *
 * Routine to handle MODFIELD commands.
 *
 */

void processModNumField(char *cmdData)
{
   int          fldnum=0;
   int          lngth=0;
   int          nextPos;
   CICS_3270_Field_t    fld;

   /* Get the row and column from the command data */
   fldnum = parseNumber(cmdData,&nextPos);

   /* Point to the new data */
   cmdData += nextPos;

   /* get the field location */
   CICS_3270_GetFieldByNum(fldnum, &fld);

   if (fld.row != -1)
   {
      /* Get the length of the new data */
      lngth = strlen(cmdData);

      if ((fld.attribute & CICS_3270_ATTR_PROT) == 0)
      {
         /* Modify the specified field, using C3270.C */
         CICS_3270_ModifyField(fld.row,fld.column,cmdData,lngth);
          }
      else
      {
         CICS_3270_Write_Log_Num(lineNum, "Attempt to modify protected field");
      }
   }
   else
   {
      CICS_3270_Write_Log_Num(lineNum, "Invalid field number");
   }
}

/*
 *
 * Routine to handle a CLEAR command.
 *
 */

void processClear()

{
   /* Write a line to the trace file */
   CICS_3270_WriteTraceLine("CLEAR COMMAND BEING PROCESSED\n");

   /* Clear the 3270 screen buffer */
   CICS_3270_ClearBuffer();

   return;
}

/*
 *
 * Routine to process the command file.  Any comment lines
 * (lines which begin with a '*', '/', ''' or ';') as well any
 * blank lines are ignored.
 *
 * If the beginning of the line is not a valid command, then an
 * error message is written and the line is ignored.
 *
 * The routine will return two pointers, one to the command
 * and the other to the data.
 *
 *
 */

void processFile(FILE* cmdFile)

{
   int   notfound = 1;
   char  ch;
   char  *tempdata;
   char  templine[MAX_LINE];
   char  errline[MAX_LINE + 50];
   char  cmd[MAX_LINE];
   char  cmdData[MAX_LINE];
   

   while  (1)
   {
       (*cmd) = '\0';
       (*cmdData) = '\0';

       /* Read a line from the command file */
       if (fgets(templine, MAX_LINE, cmdFile) == NULL)
       {
          CICS_3270_WriteTraceLine("End of file reached\n");
          return;
       }

       /* increment the line number counter */
       lineNum++;

       /* Write a trace record */
       CICS_3270_WriteTraceLine(templine);

       /* Get the first character of the command line */
       ch = templine[0];

       /*
        * Check for a comment
        */
       if ((ch != '*') && (ch != ';') && (ch != '/') &&
           (ch != '\'') && (ch != '\n'))
       {
           /*
            * Not a comment, scan for a new line character (0x0A)
            * after the command and if found replace
            * with a string termination character.
            */
           tempdata = strchr(templine, '\n');
           if (tempdata != NULL)
           {
              tempdata[0] = '\0';
           }

           /* copy the command line to the passed area */
           strcpy(cmd,templine);  /* copy to command buffer */

           /*
            * Scan for the first blank
            * after the command and point to the
            * command data, if any.
            */

           tempdata = strchr(cmd, ' ');  /* find first blank */
           if (tempdata != NULL)
           {
              tempdata[0] = '\0';        /* terminate the command */
              tempdata++;                /* move past blank */

              while (tempdata[0] == ' ') /* remove additional leading blanks */
              {
                 tempdata++;
              }

              strcpy(cmdData,tempdata);  /* save the command data */

				/* Strip any trailing blanks */
				while ((strlen(cmdData) > 0) && (cmdData[strlen(cmdData) - 1] == ' '))
				{
					cmdData[strlen(cmdData) - 1] = '\0';
				}
           }

           /*
            * Check for a valid command.  If the command
            * is valid, set found to true.
            */

           /* Make sure it's upper case */
           strupr(cmd);

           /* Check if we have a valid command */
           if ((strcmp(cmd, "MODFIELD") == 0)    ||
               (strcmp(cmd, "MODNUMFLD") == 0)   ||
               (strcmp(cmd, "SETCURSOR") == 0)   ||
               (strcmp(cmd, "DUMPSCR") == 0)     ||
               (strcmp(cmd, "READRAW") == 0)     ||
               (strcmp(cmd, "PRINTFLD") == 0)    ||
               (strcmp(cmd, "PRINTNUMFLD") == 0) ||
               (strcmp(cmd, "TRACE") == 0)       ||
               (strcmp(cmd, "WRITEPW") == 0)     ||
               (strcmp(cmd, "WRITEFIELDS") == 0) ||
               (strcmp(cmd, "KEYIN") == 0)       ||
               (strcmp(cmd, "DUMPFLDS") == 0))
           {
              /* Valid command - drop out of while loop */
              notfound = 0;

              /* Check for MODFIELD command - if so, handle it */
              if (strcmp(cmd, "MODFIELD") == 0)
              {
                 notfound = 1;
                 processModField(cmdData);
              } /* endif */

              /* Check for MODNUMFLD command - if so, handle it */
              if (strcmp(cmd, "MODNUMFLD") == 0)
              {
                 notfound = 1;
                 processModNumField(cmdData);
              } /* endif */

              /* Check for KEYIN command - if so, handle it */
              if (strcmp(cmd, "KEYIN") == 0)
              {
                 notfound = 1;
                 processKeyin(cmdData);
              } /* endif */

              /* Check for SETCURSOR command - if so, handle it */
              if (strcmp(cmd, "SETCURSOR") == 0)
              {
                 notfound = 1;
                 processSetCursor(cmdData);
              } /* endif */

              /* Check for DUMPSCR command - if so, handle it */
              if (strcmp(cmd, "DUMPSCR") == 0)
              {
                 notfound = 1;
                 buildScrFileName(cmdData);
                 CICS_3270_DumpBuffer(scrnFile,ATTR_CHAR, NULL_CHAR);
              } /* endif */

              /* Check for DUMPFLDS command - if so, handle it */
              if (strcmp(cmd, "DUMPFLDS") == 0)
              {
                 notfound = 1;
                 buildScrFileName(cmdData);
                 CICS_3270_DumpFields(scrnFile);
              } /* endif */

              /* Check for READRAW command - if so, handle it */
              if (strcmp(cmd, "READRAW") == 0)
              {
                 notfound = 1;
                 readRawFile(cmdData);
              } /* endif */

              /* Check for PRINTFLD command - if so, handle it */
              if (strcmp(cmd, "PRINTFLD") == 0)
              {
                 notfound = 1;
                 processPrintField(cmdData);
              } /* endif */

              /* Check for PRINTNUMFLD command - if so, handle it */
              if (strcmp(cmd, "PRINTNUMFLD") == 0)
              {
                 notfound = 1;
                 processPrintNumField(cmdData);
              } /* endif */

              /* Check for WRITEFIELDS command - if so, handle it */
              if (strcmp(cmd, "WRITEFIELDS") == 0)
              {
                 notfound = 1;
                 processWriteFields(cmdData);
              } /* endif */

              /* Check for WRITEPW command - if so, handle it */
              if (strcmp(cmd, "WRITEPW") == 0)
              {
                 notfound = 1;
                 if (processWritePW(cmdData, "", "") > 0)
				 {
					 CICS_3270_Write_Log_Num(lineNum, "unable to process catalog file");
				 }
              } /* endif */

              /* Check for TRACE command - if so, handle it */
              if (strcmp(cmd, "TRACE") == 0)
              {
                 notfound = 1;
                 processTrace(cmdData);
              } /* endif */
            }
                    else
            {
               /* Invalid command - tell the user and ignore */
               sprintf(errline,"Invalid Command = %s\n", cmd);
               CICS_3270_Write_Log_Num(lineNum, errline);
            } /* endif */

       }  /* endif - check for comment */

   } /* endwhile */


   /*
    * Check if we found a command.  If found, return 0,
    * otherwise, nothing more to do, so return a 1 to
    * end the program.
    */

	return;
}

void writeHelpInfo(char *cmd)

{
   printf("\n3270 File Simulator V1.1\n");
   printf("Usage: %s cmdfile <region|* <term model>\n\n", cmd);
   printf("For trace information, set DEBUG to the name of\n");
   printf("the desired trace file, which will be created or\n");
   printf("overwritten.  To display the trace information on the\n");
   printf("console, set DEBUG=stdout.  Lines beginning with a\n");
   printf("*, ;, ' or / are treated as comments.\n");
   printf("Valid commands are (<> indicates optional):\n\n");
   printf("KEYIN [keystroke data]\n");
   printf("MODFIELD row column data or ");
   printf("MODNUMFLD field-number data\n");
   printf("SETCURSOR row col\n");
   printf("CLEAR\n");
   printf("DUMPSCR  <filename>\n");
   printf("DUMPFLDS <filename>\n");
   printf("PRINTFLD row column or ");
   printf("PRINTNUMFLD field-number\n");
   printf("TRACE OFF | ON <filename>\n");
   printf("WRITEFIELDS filename\n");
   printf("WRITEPW filename <prefix>\n");
   return;
}

int main(int argc, char *argv[])

{
   FILE         *cmdFile;			/* command file */
   int			termModel=2;		/* default to 3270 model 2 */
   char			*filename;
   char			errline[MAX_LINE + 50];

   /**************************************************************/
   /* Check that if a file name was specified                    */
   /**************************************************************/

   if (argc < 2)
   {
      writeHelpInfo(argv[0]);
      return 1;
   }

   filename = argv[1];

   /**************************************************************/
   /* Check for a help request                                   */
   /**************************************************************/


   if ((filename[0] == '?') || (filename[1] == '?'))
   {
      writeHelpInfo(argv[0]);
      return 1;
   } /* endif */

   /**************************************************************/
   /* Check if a CICS terminal model was specified               */
   /**************************************************************/

   if (argc > 2)
   {
      termModel = atoi(argv[2]);
   }

   /**************************************************************/
   /* Open the command file.  Exit if the file cannot be         */
   /*  opened.                                                   */
   /**************************************************************/

   if ((cmdFile = fopen(filename, "r")) == NULL)
   {
      printf("Unable to open command file\n");
      return(100);
   }


   /**************************************************************/
   /* If debug is enabled, write out the terminal details        */
   /**************************************************************/

   sprintf(errline,"Terminal model is %d\n", termModel);
   CICS_3270_WriteTraceLine(errline);

   /**************************************************************/
   /* Initialize the 3270 data parsing module, specifying a      */
   /*  terminal model to match the installed model.              */
   /**************************************************************/

   CICS_3270_Initialize(termModel);


   /**************************************************************/
   /* Process the command file                                   */
   /**************************************************************/

   processFile(cmdFile);

   /**************************************************************/
   /* Close the command file                                     */
   /**************************************************************/

   fclose(cmdFile);
}

/*-------------*/
/* End Of File */
/*-------------*/