/*
 * Author: Jim MacNair
 * Version: 1.0
 * Date: 09/30/97
 *
 * Version 1.1.1
 * Date  05/21/98
 *
 * Description:
 * This program is a general purpose front simulator program which
 * uses the C3270.C programs to drive 3270 CICS transactions using
 * the EPI.
 *
 * The program uses an input file.  The format of the command is:
 *
 *    sim3270 input-file <cics-region|* <netname|*> <term model> >
 *
 * The input file contains command statements.  Each command
 * statement must be on a separate line.  The valid commands are:
 *
 *   BEGINTRAN tranid <initial data>
 *   MODFIELD row column data
 *   MODNUMFLD field-number data
 *   KEYIN [keystroke-data]
 *   INPUT <aid>
 *   SETCURSOR row column
 *   CLEAR
 *   WAITATI
 *   PRINTFLD row column <filename>
 *   PRINTNUMFLD field-number data
 *   DUMPSCR  <filename>
 *   DUMPFLDS <filename>
 *   DUMPRAW  filename
 *   READRAW  filename
 *   WRITEFIELDS <filename>
 *   WRITEPW  filename <prefix>
 *   TRACE OFF | ON <filename>
 *
 * BEGINTRAN is used to start a transaction.  MODFIELD is used to
 * alter a field in the screen image.
 *
 * CLEAR is a shorthand for KEYIN with predefined aid.
 *
 * The DUMPSCR and DUMPFLDS are used to write the current contents
 * of the screen to a file.  If no file is specified, then the
 * output is sent to stdout.
 *
 * When a transaction ends, the input file will be processed until
 * the next BEGINTRAN, INPUT or CLEAR statement is found.  If
 * a next transaction id was specified, then it will be used
 * and the INPUT or CLEAR will be used to send the initial input.
 * If no next transaction id was specified by the previous
 * transaction, then a BEGINTRAN is required.  Any MODFIELD,
 * DUMPSCR, DUMPFLDS or PRINTFLD statements encountered will
 * be processed.
 *
 * When a converse event is received, the input file will be
 * processed until an INPUT or CLEAR statement is found.  Any
 * BEGINTRAN statements will be skipped.  Thus, it is important
 * that the input file reflect the actual transaction flow.
 * Again, any MODFIELD, DUMPSCR, DUMPFLDS or PRINTFLD
 * statements encountered will be processed.
 *
 * The PRINTFLD, DUMPSCR and DUMPFLDS write out information
 * about the current 3270 screen image.  PRINTFLD will write
 * the current begining location (row and column), attribute,
 * data length and actual data for the field which contains
 * the given row and column to stdout.  DUMPSCR will write
 * the current screen image to a file, substituting an @
 * character for each attribute byte.  DUMPFLDS will write
 * one line with the location and characteristics of each
 * attribute byte to a file.
 *
 * The WRITEFIELDS statement will write a description of
 * each field on a separate line.  The line contains the
 * location (row and column), length and up to 30 characters 
 * of data for each field, plus the basic field attributes.
 *
 * The WRITEPW statement will update an Interspace (TM)
 * catalog file with information about the current screen.
 * Only unprotected fields are added to the catalog.  If
 * an unprotected field is preceded on the same line by
 * a protected field, then this is likely to be a label
 * and an attempt is made to use this label as a field
 * name.  Otherwise, a variable name of the format 
 * FIELDnnn is assigned.  This command will update the
 * catalog if it exists, and will create a new catalog
 * if the cataog does not exist.  It will rename the 
 * existing catalog to a file extension of "bak", and
 * the updated catalog will retain the extension of
 * "cat".  
 *
 * The DUMPRAW command will write the 3270 data stream itself
 * to a file.  The datastream will include the write command
 * and the write control character (WCC) in the first two
 * positions, followed by the orders and data.  This command
 * makes it easy to capture an actual 3270 datastream.  This
 * command is aimed at customers who want to feed the datastream
 * to another program.  The filename will be saved and used the
 * next time a 3270 datastream is received.  When a datastream is
 * received, the raw file name will be reset.  If the output file
 * already exists, it will be overwritten with the contents of
 * the 3270 datastream.
 *
 * WAITATI will enable ATI processing, and then return to
 * the main event handler loop.  Processing of the
 * command file will be suspended until a host initiated
 * transaction begins, at which point processing of the
 * command file will continue.
 *
 * Lines beginning with a "*", "'", "/" or ";" will be
 * treated as comments and ignored.
 *
 * Extra trace information will be written if the
 * DEBUG environment variable is set.  If used, the DEBUG
 * variable should be set to the name of the desired DEBUG
 * file.  To direct trace information to stdout, set
 * DEBUG=stdout.
 *
 * Command lines can be a maximum of 256 characters long.  Lines longer
 * than this will be truncated.
 *
 * This program is written for the Windows/NT environment, but should
 * be portable to other platforms with little or no change.
 *
 */

/*
 *
 * Updates: V1.11 - Bug fix
 *   1) Fixed bug with if statement for netname and region name
 *
 * Updates: V1.1 - Bug fixes
 *
 *   1) Fixed bug in handling of Repeat to Address order
 *   2) Fixed bug in handling of Program Tab order
 *
 * Updates: V1.1 - enhancements
 *
 *   1) Added support for writing field location table
 *   2) Added support for creating Planetworks Interspace (TM) catalog
 *   3) Added support for dumping raw 3270 data stream to a file
 *
 */

#define CICS_W32
#define u_short_t cics_ushort_t
#define u_byte_t  cics_ubyte_t
#include <cics_epi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "c3270.h"
#include "epimsg.h"
#include "errorlog.h"

static char Version[]=
"@(#)SIM3270 V1.11 - 3270 EPI Simulator - Jim MacNair ("__DATE__")";

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#define MAX_LINE   255          /* Maximum length of command */
#define ATTR_CHAR  '@'          /* default screen dump attribute char */
#define NULL_CHAR  '_'          /* default screen dump null char */
#define SCREEN_FILE_NAME   "scrfile.txt\0"
#define NEW_TRAN_DATA "\x027  \0"
#define PWCAT_EXT ".cat"
#define TEMP_EXT  ".tmp"
#define BAK_EXT   ".bak"

/*
 * Terminal index used by EPI
 */
   u_short_t    TermIndex=CICS_EPI_TERM_INDEX_NONE;

/*
 * Command File
 */

   FILE         *cmdFile;

/*
 * Line number within the command File
 */

   int          lineNum=0;

/*
 * Transaction active indicator
 *
 * This indicator is primarily used to determine if a
 * clear must be sent to the host or not.  If no transaction
 * is present, then the clear command merely clears the local
 * buffer.  If a transaction is running, then it should be
 * sent in response to a converse.
 *
 */

   int          inTran=0;

/*
 * Indicates if ATI starts have been enabled
 */

   int          ATIEnabled=0;

/*
 * Name of Default Screen Dump File
 */

   char         scrnFile[255]="\0";;

/*
 * Name of Default Raw Buffer Dump File
 */

   char         rawFile[255]="\0";;

/*
 *
 * Routine to check if a file name is specified for a
 * screen dump file.  If it is, then that name is used,
 * otherwise the default name defined above is used.
 *
 */
void buildScrFileName(char *cmdData)

{
   if ((cmdData == NULL) || ((*cmdData) == '\0'))
   {
      strcpy(scrnFile,SCREEN_FILE_NAME);
   }
   else
   {
      strcpy(scrnFile,cmdData);
   } /* endif */
}

/*
 *
 * Routine to check if a file name is specified for a
 * 3270 dump file.  If it is, then that name is used,
 * otherwise the default name defined above is used.
 *
 */
void buildRawFileName(char *cmdData)

{
   if ((cmdData == NULL) || ((*cmdData) == '\0'))
   {
	   /* no file name specified */
      rawFile[0] = 0;
   }
   else
   {
      strcpy(rawFile,cmdData);
   } /* endif */
}

/*
 *
 * Routine to find the next variable in the command line
 * and terminate, moving the data pointer.
 *
 */

void parseVar(char *cmdData, int *nextPos)
{
   char         *newData;

   /*
    * Skip over any leading blanks
    */

   while (cmdData[0] == ' ')
   {
      cmdData++;
   } /* endwhile */

   if ((cmdData != NULL) && ((*cmdData) != '\0'))
   {
      newData = strchr(cmdData, ' ');  /* find trailing blank */
      if (newData != NULL)
      {
         newData[0] = '\0';  /* Terminate the string */
      } /* endif */

      newData++;             /* move the pointer */
      (*nextPos) = (int) (newData - cmdData);
   } /* endif */

   return;
}

/*
 *
 * Routine to handle an escape sequence for an 3270
 * attention key.
 *
 */

int processAID(char *cmdData)
{
   int          i=0;
   char         ch1;
   char         ch2;

   /* translate the escape sequence characters to upper case */
   strupr(cmdData);

   ch1 = cmdData[1];
   ch2 = cmdData[2];

   switch (ch1)
   {
       case '0':
          if ((ch2 >= '1') && (ch2 <= '9'))
          {
             i = CICS_3270_AID_PF1 + (ch2 - '1');  /* PF1-9 */
          } /* endif */
          break;

       case '1':
          switch (ch2)
          {
              case '0':
                 i = CICS_3270_AID_PF10;  /* PF10 */
                 break;

              case '1':
                 i = CICS_3270_AID_PF11;  /* PF11 */
                 break;

              case '2':
                 i = CICS_3270_AID_PF12;  /* PF12 */
                 break;

              case '3':
                 i = CICS_3270_AID_PF13;  /* PF13 */
                 break;

              case '4':
                 i = CICS_3270_AID_PF14;  /* PF14 */
                 break;

              case '5':
                 i = CICS_3270_AID_PF15;  /* PF15 */
                 break;

              case '6':
                 i = CICS_3270_AID_PF16;  /* PF16 */
                 break;

              case '7':
                 i = CICS_3270_AID_PF17;  /* PF17 */
                 break;

              case '8':
                 i = CICS_3270_AID_PF18;  /* PF18 */
                 break;

              case '9':
                 i = CICS_3270_AID_PF19;  /* PF19 */
                 break;

              default:
                 i = 0;
          } /* endswitch */

          break;

       case '2':
          switch (ch2)
          {
              case '0':
                 i = CICS_3270_AID_PF20;  /* PF20 */
                 break;

              case '1':
                 i = CICS_3270_AID_PF21;  /* PF21 */
                 break;

              case '2':
                 i = CICS_3270_AID_PF22;  /* PF22 */
                 break;

              case '3':
                 i = CICS_3270_AID_PF23;  /* PF23 */
                 break;

              case '4':
                 i = CICS_3270_AID_PF24;  /* PF24 */
                 break;

              default:
                 i = 0;
          } /* endswitch */

          break;

       case 'A':
          switch (ch2)
          {
              case '1':
                 i = CICS_3270_AID_PA1;  /* PA1 */
                 break;

              case '2':
                 i = CICS_3270_AID_PA2;  /* PA2 */
                 break;

              case '3':
                 i = CICS_3270_AID_PA3;  /* PA3 */
                 break;

              default:
                 i = 0;
          } /* endswitch */

          break;

       case 'C':
          if (ch2 == 'L')
          {
             i = CICS_3270_AID_CLEAR;       /* Clear */
          }
          else
          {
             if (ch2 == 'S')
             {
                i = CICS_3270_AID_PEN;      /* Cursor Select */
             } /* endif */
          } /* endif */

          break;

       case 'E':
          if (ch2 == 'N')
          {
             i = CICS_3270_AID_ENTER;       /* Enter */
          }

          break;

       case 'M':
          if (ch2 == 'E')
          {
             i = CICS_3270_AID_MSRE;       /* End secure MSR */
          }

          break;

       case 'O':
          if (ch2 == 'P')
          {
             i = CICS_3270_AID_OPID;       /* Operator ID */
          }

          break;

       case 'T':
          if (ch2 == 'R')
          {
             i = CICS_3270_AID_TRIG;       /* Trigger Field */
          }

          break;

       default:
          i = 0;
   } /* endswitch */

   return i;
}

/*
 *
 * Routine to find the next number in the command line
 * and return it as an integer, moving the data pointer.
 *
 */

int parseNumber(char *cmdData, int *nextPos)
{
   int          i=0;
   char         *newData;

   /* initialize next position offset to 0, in case */
   /* there is no parameter to process              */
   *nextPos = 0;

   /*
    * Skip over any leading blanks
    */

   while (cmdData[0] == ' ')
   {
      cmdData++;
   } /* endwhile */

   /*
    * Find the end of the number
    */

   if ((cmdData != NULL) && ((*cmdData) != '\0'))
   {
      newData = strchr(cmdData, ' ');  /* find trailing blank */
      if (newData != NULL)
      {
         newData[0] = '\0';  /* Terminate the string */
      }
      else
      {
         newData = cmdData + strlen(cmdData) - 1;
      } /* endif */

      if ((cmdData[0] >= '0') && (cmdData[0] <= '9'))
      {
         i = atoi(cmdData);     /* get the value */
      }
      else
      {
         if (cmdData[0] == '&')
         {
            i = processAID(cmdData);
         } /* endif */
      } /* endif */

      newData++;             /* move the pointer */
      (*nextPos) = (int) (newData - cmdData);
   } /* endif */

   return i;
}

/*
 *
 * Routine to handle TRACE commands.
 *
 */

void processTrace(char *cmdData)
{
   int                 rc;
   char               *fileName;
   char               *endfileName;

   /* Skip leading blanks in cmdData */
   while (cmdData[0] == ' ')
   {
           cmdData++;
   }

   /* Find the end of the first parameter */
   fileName = strchr(cmdData,' ');
   if (fileName != NULL)
   {
      fileName[0] = '\0';   /* terminate the first parm */
      fileName++;           /* move past string terminator */

      while (fileName[0] == ' ')
      {
         fileName++;  /* skip leading blanks */
      }

      /* look for trailing  blanks */
      endfileName = strchr(fileName, ' ');
      if (endfileName != NULL)
      {
         endfileName[0] = '\0';
      } /* endif */
   }

   /* Now set a variable to indicate if we are */
   /* turning trace on or off or querying status */

   if (strcmp(cmdData,"ON\0") == 0)
   {
      rc = CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_ON, fileName);
   }
   else
   {
      if (strcmp(cmdData,"APPEND\0") == 0)
      {
         rc = CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_APPEND, fileName);
      }
      else
      {
         if (strcmp(cmdData,"OFF\0") == 0)
         {
            rc = CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_OFF, NULL);
         }
         else
         {
            rc = CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_INQ, NULL);
         }
      }
   }

 if (rc)
 {
    printf("Trace is enabled\n");
 }
 else
 {
    printf("Trace is off\n");
 }
}

/*
 *
 * Routine to handle SETCURSOR commands.
 *
 */

void processSetCursor(char *cmdData)

{
   int                 row=0;
   int                 col=0;
   int                 nextPos;

   /* Get the row and column from the command data */
   row = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;

   /* skip over any blanks */
   while (*cmdData == ' ')
   {
      cmdData++;
   }

   if (strlen(cmdData) > 0)
   {
      col = parseNumber(cmdData,&nextPos);

      /* Print the characteristics of the field */
      /* containing the desired location */
      CICS_3270_SetCursor(row, col);
   }
   else
   {
      CICS_3270_Write_Log_Num(lineNum,"missing or invalid parameters for SETCURSOR");
   }
}

/*
 *
 * Routine to handle READRAW commands.
 *
 *
 * A file containing a complete 3270 data stream is read from a
 * file and then applied to the input buffer.  The file must begin
 * with a valid write command followed by a write control character.
 * The file cannot exceed 32767 characters.
 *
 */

void readRawFile(char *cmdData)

{
	FILE *inpfile;
	char errline[255];
	char databuffer[32768];
	int  len;

	/* open the file for binary input */
	inpfile = fopen(cmdData, "rb");
	if (inpfile == NULL)
	{
		/* write an error message to the log and exit */
		sprintf(errline, "unable to open 3270 input file - %s", cmdData);
		CICS_3270_Write_Log_Num(lineNum,errline);
		return;
	}

	/* read the file into the buffer */
	len = fread(databuffer, sizeof(char), sizeof(databuffer), inpfile);

	CICS_3270_Parse(databuffer, len);

	/* close the raw input file */
	fclose(inpfile);
}

/*
 *
 * Routine to handle WRITEFIELDS commands.
 *
 */

void processWriteFields(char *cmdData)

{
	/* write a file containing a line per field */
	CICS_3270_WriteFields(cmdData);
}

/*
 *
 * Routine to create an empty PlanetWorks catalog file.
 *
 */

int createPWcat(char *fileName)

{
	FILE    *catfile;

	/* create the catalog file */
	catfile = fopen(fileName, "w");
	if (catfile == NULL)
	{
		/* unable to create file */
		return 100;
	}

	/* Write some comments to identify the file */
	fprintf(catfile, "#\n");
	fprintf(catfile, "# Interspace service definition file for CICS 3270\n");
	fprintf(catfile, "# Captured by Sim3270 program from IBM SupportPac CA04\n");
	fprintf(catfile, "#\n");
	fprintf(catfile, "# Fields are defined here\n");
	fprintf(catfile, "# NAME		TYPE	MEMORY	DISPLAY	DESCRIPTION\n");

	/* Indicate the beginning of the field definitions */
	fprintf(catfile, "*FIELDS\n");

	/* insert a service definition for the transaction */
	fprintf(catfile, "#\n");
	fprintf(catfile, "# Service Definitions\n");
	fprintf(catfile, "#\n");
	fprintf(catfile, "*SERVICES\n");
	fprintf(catfile, "#\n");

	/* Close the file */
	fclose(catfile);

	/* indicate normal return */
	return 0;
}

/*
 *
 * Routine to handle WRITEPW (write interspace catalog files) commands.
 *
 */

int processWritePW(char *cmdData, char *trnid, char *syst)

{
	FILE *oldcat;
	FILE *newcat;
	char filename[255];
	char newfilename[255];
	char bakfilename[255];
	char pref[255];
	char errline[255];
	char *prefix;
	char *fileext;
	int  foundbak = 0;
	int  rc;

	/* get the catalog file name */
	strcpy(filename, cmdData);

	/* look for a blank following the file name */
	prefix = strchr(filename, ' ');

	/* did we find a blank? */
	if (prefix == NULL)
	{
		/* point to a null prefix if none is specified */
		prefix = "";
	}
	else
	{
		/* terminate the file name string */
		prefix[0] = '\0';

		/* skip the null character */
		prefix++;

		/* skip any leading blanks */
		while (prefix[0] == ' ')
		{
			prefix++;
		}
	}

	/* save the prefix */
	strcpy(pref, prefix);

	/* Parse the catalog file name, and remove the extension, if any */
	fileext = strchr(filename, '.');

	/* did we find one? */
	if (fileext != NULL)
	{
		/* remove the file extension */
		fileext[0] = '\0';
	}

	/* get the new catalog name */
	strcpy(newfilename, filename);
	strcpy(bakfilename, filename);

	/* append the right file extensions for the old and new catalogs */
	strcat(filename, PWCAT_EXT);
	strcat(newfilename, TEMP_EXT);
	strcat(bakfilename, BAK_EXT);

	/* try to open the catalog, to see if it already exists */
	oldcat = fopen(filename, "r");
	if (oldcat == NULL)
	{
		/* Check for a .bak file */
		oldcat = fopen(bakfilename, "r");
		if (oldcat == NULL)
		{
			/* couldn't open the file, let's try to create one */
			if (createPWcat(filename) == 0)
			{
				/* now, try to open it again */
				oldcat = fopen(filename, "r");
				if (oldcat == NULL)
				{
					/* unable to open the catalog file */
					return 100;
				}
			}
			else
			{
				/* unable to create catalog file */
				return 100;
			}
		}
		else
		{
			/* found a file with a .bak extension */
			foundbak = 1;
		}
	}

	/* try to open the temporary new catalog */
	newcat = fopen(newfilename, "w");
	if (newcat == NULL)
	{
		/* close the old catalog */
		fclose(oldcat);

		/* problem opening the temporary new catalog file */
		return 100;
	}

	/* write a Planetworks Interspace(R) catalog file */
	CICS_3270_WritePW(oldcat, newcat, pref, trnid, syst);

	/* close the two files */
	fclose(oldcat);
	fclose(newcat);

	/* Lastly, rename the files, leaving the old catalog */
	/* with an extension of .bak                         */
	/* Do this in several steps, first erasing any old   */
	/* .bak files, then renaming the current file to     */
	/* .bak, and finally renaming the temporary file     */
	/* to an .cat extension.                             */

	/* check if we couldn't find the .cat file and used the .bak file */
	if (foundbak == 0)
	{
		/* delete the old .bak file */
		rc = remove(bakfilename);
		if ((rc == 0) || (rc = -1))
		{
			/* delete worked or no such file, now try to rename the .cat file to .bak */
			rc = rename(filename, bakfilename);
			if (rc != 0)
			{
				/* can't rename the original file - return error */
				sprintf(errline, "unable to rename %s to %s - %d\n", filename, bakfilename, rc);

				/* write message to error log */
				CICS_3270_Write_Log_Num(lineNum, errline);

				return 100;
			}
		}
		else
		{
			/* can't erase the old .bak file - return error */
			sprintf(errline, "unable to delete the file %s - %d\n", bakfilename, rc);

			/* write message to error log */
			CICS_3270_Write_Log_Num(lineNum, errline);

			return 100;
		}
	}

	/* rename the .tmp file to .cat to finish the operation */
	rc = rename(newfilename, filename);

	/* check if the rename worked */
	if (rc != 0)
	{
		/* can't rename the new file to the correct extension */
		sprintf(errline, "unable to rename %s to %s - %d\n", newfilename, filename, rc);

		/* write message to error log */
		CICS_3270_Write_Log_Num(lineNum, errline);

		return 100;
	}

	return 0;
}

/*
 *
 * Routine to handle PRINTNUMFLD commands.
 *
 */

void processPrintNumField(char *cmdData)

{
   int     row=0;
   int     col=0;
   int     num=0;
   int     nextPos;

   /* Get the field number from the command data */
   num = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;
   parseVar(cmdData, &nextPos);

   /* get the row and column of the field */
   CICS_3270_GetFieldLocByNum(num, &row, &col);

   if (row != -1)
   {
      /* Print the characteristics of the field */
      /* containing the desired location */
      CICS_3270_PrintField(row, col, cmdData);
   }
   else
   {
      CICS_3270_Write_Log_Num(lineNum, "Invalid print field number");
   }
}

/*
 *
 * Routine to handle PRINTFIELD commands.
 *
 */

void processPrintField(char *cmdData)
{
   int                 row=0;
   int                 col=0;
   int                 nextPos;

   /* Get the row and column from the command data */
   row = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;
   col = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;
   parseVar(cmdData, &nextPos);

   /* Print the characteristics of the field */
   /* containing the desired location */
   CICS_3270_PrintField(row, col, cmdData);
}

/*
 *
 * Routine to handle KEYIN commands.
 *
 */

void processKeyin(char *cmdData)
{
   /* Process the keyin request */
   CICS_3270_Keyin(cmdData);
}

/*
 *
 * Routine to handle MODFIELD commands.
 *
 */

void processModField(char *cmdData)
{
   int          row=0;
   int          col=0;
   int          lngth=0;
   int          nextPos;

   /* Get the row and column from the command data */
   row = parseNumber(cmdData,&nextPos);
   cmdData += nextPos;
   col = parseNumber(cmdData,&nextPos);
   /* Point to the new data */
   cmdData += nextPos;

   /* Get the length of the new data */
   lngth = strlen(cmdData);

   /* Modify the specified field, using C3270.C */
   CICS_3270_ModifyField(row,col,cmdData,lngth);
}

/*
 *
 * Routine to handle MODFIELD commands.
 *
 */

void processModNumField(char *cmdData)
{
   int          fldnum=0;
   int          lngth=0;
   int          nextPos;
   CICS_3270_Field_t    fld;

   /* Get the row and column from the command data */
   fldnum = parseNumber(cmdData,&nextPos);

   /* Point to the new data */
   cmdData += nextPos;

   /* get the field location */
   CICS_3270_GetFieldByNum(fldnum, &fld);

   if (fld.row != -1)
   {
      /* Get the length of the new data */
      lngth = strlen(cmdData);

      if ((fld.attribute & CICS_3270_ATTR_PROT) == 0)
      {
         /* Modify the specified field, using C3270.C */
         CICS_3270_ModifyField(fld.row,fld.column,cmdData,lngth);
          }
      else
      {
         CICS_3270_Write_Log_Num(lineNum, "Attempt to modify protected field");
      }
   }
   else
   {
      CICS_3270_Write_Log_Num(lineNum, "Invalid field number");
   }
}

/*
 *
 * Routine to move to the next command line.  Any comment lines
 * (lines which begin with a '*', '/', ''' or ';') as well any
 * blank lines are ignored.
 *
 * If the beginning of the line is not a valid command, then an
 * error message is written and the line is ignored.
 *
 * The routine will return two pointers, one to the command
 * and the other to the data.
 *
 *
 */

int nextLine(char *cmd, char *cmdData, char *trnid, char *syst)

{
   int   notfound = 1;
   char  ch;
   char  *tempdata;
   char  templine[MAX_LINE];
   char  errline[MAX_LINE + 50];

   while  (notfound)
   {
       (*cmd) = '\0';
       (*cmdData) = '\0';

       /* Read a line from the command file */
       if (fgets(templine, MAX_LINE, cmdFile) == NULL)
       {
          CICS_3270_WriteTraceLine("End of file reached\n");
          return 1;
       }

       /* increment the line number counter */
       lineNum++;

       /* Write a trace record */
       CICS_3270_WriteTraceLine(templine);

       /* Get the first character of the command line */
       ch = templine[0];

       /*
        * Check for a comment
        */
       if ((ch != '*') && (ch != ';') && (ch != '/') &&
           (ch != '\'') && (ch != '\n'))
       {
           /*
            * Not a comment, scan for a new line character (0x0A)
            * after the command and if found replace
            * with a string termination character.
            */
           tempdata = strchr(templine, '\n');
           if (tempdata != NULL)
           {
              tempdata[0] = '\0';
           }

           /* copy the command line to the passed area */
           strcpy(cmd,templine);  /* copy to command buffer */

           /*
            * Scan for the first blank
            * after the command and point to the
            * command data, if any.
            */

           tempdata = strchr(cmd, ' ');  /* find first blank */
           if (tempdata != NULL)
           {
              tempdata[0] = '\0';        /* terminate the command */
              tempdata++;                /* move past blank */

              while (tempdata[0] == ' ') /* remove additional leading blanks */
              {
                 tempdata++;
              }

              strcpy(cmdData,tempdata);  /* save the command data */

				/* Strip any trailing blanks */
				while ((strlen(cmdData) > 0) && (cmdData[strlen(cmdData) - 1] == ' '))
				{
					cmdData[strlen(cmdData) - 1] = '\0';
				}
           }

           /*
            * Check for a valid command.  If the command
            * is valid, set found to true.
            */

           /* Make sure it's upper case */
           strupr(cmd);

           /* Check if we have a valid command */
           if ((strcmp(cmd, "MODFIELD") == 0)    ||
               (strcmp(cmd, "MODNUMFLD") == 0)   ||
               (strcmp(cmd, "BEGINTRAN") == 0)   ||
               (strcmp(cmd, "INPUT") == 0)       ||
               (strcmp(cmd, "CLEAR") == 0)       ||
               (strcmp(cmd, "SETCURSOR") == 0)   ||
               (strcmp(cmd, "DUMPSCR") == 0)     ||
               (strcmp(cmd, "DUMPRAW") == 0)     ||
               (strcmp(cmd, "READRAW") == 0)     ||
               (strcmp(cmd, "PRINTFLD") == 0)    ||
               (strcmp(cmd, "PRINTNUMFLD") == 0) ||
               (strcmp(cmd, "TRACE") == 0)       ||
               (strcmp(cmd, "WRITEPW") == 0)     ||
               (strcmp(cmd, "WRITEFIELDS") == 0) ||
               (strcmp(cmd, "KEYIN") == 0)       ||
               (strcmp(cmd, "WAITATI") == 0)     ||
               (strcmp(cmd, "DUMPFLDS") == 0))
           {
              /* Valid command - drop out of while loop */
              notfound = 0;

              /* Check for MODFIELD command - if so, handle it */
              if (strcmp(cmd, "MODFIELD") == 0)
              {
                 notfound = 1;
                 processModField(cmdData);
              } /* endif */

              /* Check for MODNUMFLD command - if so, handle it */
              if (strcmp(cmd, "MODNUMFLD") == 0)
              {
                 notfound = 1;
                 processModNumField(cmdData);
              } /* endif */

              /* Check for KEYIN command - if so, handle it */
              if (strcmp(cmd, "KEYIN") == 0)
              {
                 notfound = 1;
                 processKeyin(cmdData);
              } /* endif */

              /* Check for SETCURSOR command - if so, handle it */
              if (strcmp(cmd, "SETCURSOR") == 0)
              {
                 notfound = 1;
                 processSetCursor(cmdData);
              } /* endif */

              /* Check for DUMPSCR command - if so, handle it */
              if (strcmp(cmd, "DUMPSCR") == 0)
              {
                 notfound = 1;
                 buildScrFileName(cmdData);
                 CICS_3270_DumpBuffer(scrnFile,ATTR_CHAR, NULL_CHAR);
              } /* endif */

              /* Check for DUMPFLDS command - if so, handle it */
              if (strcmp(cmd, "DUMPFLDS") == 0)
              {
                 notfound = 1;
                 buildScrFileName(cmdData);
                 CICS_3270_DumpFields(scrnFile);
              } /* endif */

              /* Check for DUMPRAW command - if so, handle it */
			  /* Save the raw file name for the next input buffer */
              if (strcmp(cmd, "DUMPRAW") == 0)
              {
                 notfound = 1;
                 buildRawFileName(cmdData);
              } /* endif */

              /* Check for READRAW command - if so, handle it */
              if (strcmp(cmd, "READRAW") == 0)
              {
                 notfound = 1;
                 readRawFile(cmdData);
              } /* endif */

              /* Check for PRINTFLD command - if so, handle it */
              if (strcmp(cmd, "PRINTFLD") == 0)
              {
                 notfound = 1;
                 processPrintField(cmdData);
              } /* endif */

              /* Check for PRINTNUMFLD command - if so, handle it */
              if (strcmp(cmd, "PRINTNUMFLD") == 0)
              {
                 notfound = 1;
                 processPrintNumField(cmdData);
              } /* endif */

              /* Check for WRITEFIELDS command - if so, handle it */
              if (strcmp(cmd, "WRITEFIELDS") == 0)
              {
                 notfound = 1;
                 processWriteFields(cmdData);
              } /* endif */

              /* Check for WRITEPW command - if so, handle it */
              if (strcmp(cmd, "WRITEPW") == 0)
              {
                 notfound = 1;
                 if (processWritePW(cmdData, trnid, syst) > 0)
				 {
					 CICS_3270_Write_Log_Num(lineNum, "unable to process catalog file");
				 }
              } /* endif */

              /* Check for TRACE command - if so, handle it */
              if (strcmp(cmd, "TRACE") == 0)
              {
                 notfound = 1;
                 processTrace(cmdData);
              } /* endif */
            }
                    else
            {
               /* Invalid command - tell the user and ignore */
               sprintf(errline,"Invalid Command = %s\n", cmd);
               CICS_3270_Write_Log_Num(lineNum, errline);
            } /* endif */

       }  /* endif - check for comment */

   } /* endwhile */


   /*
    * Check if we found a command.  If found, return 0,
    * otherwise, nothing more to do, so return a 1 to
    * end the program.
    */

   if ((*cmd) == '\0')
   {
      return 1;
   }
   else
   {
      return 0;
   } /* endif */
}

/*
 *
 * Routine to dump the raw received buffer to a file.  The file name is
 * in the rawFile variable.
 *
 */

void dumpRawData(u_byte_t *data, int length)

{
	FILE	*raw;	/* the raw file to hold the buffer */
	char	msgline[1024];

	if ((raw = fopen(rawFile, "wb")) != NULL)
	{
		/* record the event */
		sprintf(msgline, "Dumping raw buffer to file - %s\n", rawFile);
        CICS_3270_WriteTraceLine(msgline);

		/* write the raw data to the file */
		fwrite(data, 1, length, raw);

		/* close the file */
		fclose(raw);
	}
	else
	{
		/* error opening file */
		sprintf(msgline, "Unable to open raw file - %s\n", rawFile);
        CICS_3270_WriteTraceLine(msgline);
	}
}

/*
 *
 * Handle outbound 3270 data.  This function is called when data is received
 * from CICS.  The internal model of the 3270 terminal is updated with the
 * data.  It is also called when a transaction ends.
 *
 */

void HandleOutbound(u_byte_t *data, int length)

{
        CICS_3270_Parse((char *)data, length);
        return;
}

/*
 *
 * Routine to issue an EPI begin transaction.  If a
 * transaction is successfully started, then it returns
 * a 0, otherwise a 1 is returned.
 *
 */

int epiBeginTran(char *transid, char *inputdata, int lngth)

{
   int          rc;
   char         errmsg[CICS_3270_MAX_EPI_MSG_LEN];
   char         errline[MAX_LINE + 50];

   if (transid != '\0')
   {
      rc = CICS_EpiStartTran(
                      TermIndex,
                      transid,
                      inputdata,
                      (cics_ushort_t) lngth);

      if (rc == CICS_EPI_NORMAL)
      {
         inTran = 1;
         sprintf(errline, "Transaction (%s) started\n", transid);
         CICS_3270_WriteTraceLine(errline);
      }
      else
      {
         /* Log the error */
         sprintf(errline, "Transaction (%s) failed to start\n", transid);
         CICS_3270_WriteTraceLine(errline);

         CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
         sprintf(errline, "%d: CICS_EpiStartTran %s\n", rc,errmsg);
         CICS_3270_Write_Log_Num(lineNum,errline);
         CICS_3270_WriteTraceLine(errline);
      }
   }
   else
   {
      /* set dummy return code to indicate error */
      rc = 100;

      /* Log the error */
      sprintf(errline, "Transid missing - nothing started\n", rc,errmsg);
      CICS_3270_Write_Log_Num(lineNum,errline);
   } /* endif */

   return rc;
}

/*
 *
 * Routine to write an error message to the log and trace if there
 * is no input data with a CICS transaction code.
 *
 */

void noTransactionData()

{
   char    errline[MAX_LINE + 50];

   sprintf(errline, "No input data for transaction code\n");
   CICS_3270_Write_Log_Num(lineNum, errline);
   CICS_3270_WriteTraceLine(errline);
}

/*
 *
 * Routine to extract a transaction id from the input data
 *
 */

int parseTransid(char *cmdData, int lngth, char *trnid)

{
   int     i;           /* work variable   */
   int     maxchar;     /* Characters in input datastream */

   /* get the length of the input data */
   maxchar = lngth;

   /* limit the maximum number of characters */
   if (maxchar > CICS_EPI_TRANSID_MAX)
   {
      /* but no more than TRANSID */
      maxchar = CICS_EPI_TRANSID_MAX;
   } /* endif */

   /* pad transaction id with blanks */
   memset(trnid, ' ', CICS_EPI_TRANSID_MAX);
   trnid[CICS_EPI_TRANSID_MAX] = '\0';

   /* copy in up to maxchar characters */
   /* skipping the aid and cursor position */
   i = 0;
   while (i < maxchar)
   {
      /* check for 3270 order or a space */
      if (cmdData[i] > ' ')
      {
         /* move a character to the transaction id */
         trnid[i] = cmdData[i];
      }
      else
      {
         /* stop if we hit a 3270 SBA order */
         break;
      } /* endif */

      i++;
   } /* endwhile */

   /* indicate the number of characters used in the transaction id */
   return i;
}

/*
 *
 * Routine to handle an INPUT command.  The command data
 *
 */

int processInput(char *cmdData, char *nextrnid)

{
   char    *data;       /* Inbound Data Stream    */
   char    trnid[CICS_EPI_TRANSID_MAX + 1]; /* transaction to start */
   int     maxchar=0;   /* Characters in input datastream */
   int     lngth;       /* Data Length            */
   int     nextPos;     /* Bytes for this parameter */
   char    *transid;    /* Pointer to transaction id in data stream */
   char    aid;         /* 3270 aid byte   */
   int     rc;          /* EPI return code */
   char    errmsg[CICS_3270_MAX_EPI_MSG_LEN];
   char    errline[MAX_LINE + 50];

   aid = parseNumber(cmdData, &nextPos);

   /* Handle defaults if no data or invalid data */
   /* If aid byte not given, default to enter */
   /*  row and column will default to 0       */

   if (aid == 0)
   {
      aid = CICS_3270_AID_ENTER;  /* Enter key */
   } /* endif */

   /* write trace line */
   sprintf(errline,"INPUT command being processed - aid = %d - next transid(%s)\n",
               aid, nextrnid);
   CICS_3270_WriteTraceLine(errline);

   /* if this is a clear operation, we must clear the buffer */
   if (aid == CICS_3270_AID_CLEAR)
   {
      CICS_3270_ClearBuffer();

      /* check if we need to send this to the host */
      if ((inTran == 0) && ((nextrnid == NULL) || ((*nextrnid) == 0)))
      {
         /* write trace line */
         CICS_3270_WriteTraceLine("No transaction, buffer cleared, nothing sent to host\n");

         /* return dummy return code to indicate no data sent */
         return 100;
      } /* endif */
   } /* endif */

   /* Build the 3270 data stream to send in,   */
   /* from the current 3270 screen image.      */
   /* If a next transid was returned from the  */
   /* last transaction, then we will use it.   */
   /* Otherwise, we will try to find a transid */
   /* in the input data.                       */

   CICS_3270_GetInbound(aid,&data,&lngth);

   /* Check if we need to send the input to the host as a reply */
   if (inTran == 1)
   {
      /* write trace line */
      CICS_3270_WriteTraceLine("Transaction running, INPUT sent as reply\n");

      rc = CICS_EpiReply(TermIndex,
                         (u_byte_t *) data,
                         (u_short_t) lngth);

      if (rc != CICS_EPI_NORMAL)
      {
         CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
         sprintf(errline, "%d: CICS_EpiReply/Clear %s\n",
                 rc,errmsg);
         CICS_3270_Write_Log_Num(lineNum, errline);
         CICS_3270_WriteTraceLine(errline);
      }
   }
   else
   {
      /* check if next transid has been specified */
      if ((nextrnid == NULL) || ((*nextrnid) == 0))
      {
         /* At this point, no transaction is running and no next transid */
         /* has been specified.  We will attempt to start a new          */
         /* transaction, if there is data in the screen buffer to use    */
         /* for a transaction code                                       */

         /* Do we have enough data to start a transaction? */
         if (((lngth > 3) && (data[3] != CICS_3270_ORDER_SBA)) || (lngth > 6))
         {
            /* point to the input data stream beginning */
            transid = data + 3;   /* point to data              */
            maxchar = lngth - 3;  /* How many chars do we have? */

            /* see if we have to skip an SBA sequence */
            if (transid[0] == CICS_3270_ORDER_SBA)
            {
               transid += 3;   /* Skip the SBA sequence */
               maxchar -= 3;   /* Adjust the count      */
            } /* endif */

            /* extract the transaction id */
            rc = parseTransid(transid, maxchar, trnid);

            if (rc > 0)
            {
               rc = epiBeginTran(trnid, data, lngth);
            }
            else
            {
               /* write error to the error log and trace file */
               noTransactionData();
            } /* endif */
         }
         else
         {
            /* write error to the error log and trace file */
            noTransactionData();
         } /* endif */
      } /* endif */
      else
      {
         /* write trace line */
         sprintf(errline,"No transaction running, Next transid = %s\n", nextrnid);
         CICS_3270_WriteTraceLine(errline);

         /* Send the input using the EPI */
         rc =  epiBeginTran(nextrnid, data, lngth);
      } /* endif */
   } /* endif */

   /* Free the storage */
   free(data);

   /*printf("\n")*/;
   return rc;
}

/*
 *
 * Routine to handle a CLEAR command.
 *
 */

void processClear(char *nextTran)

{
   char    *cmdData;    /* Aid byte character */

   /* Write a line to the trace file */
   CICS_3270_WriteTraceLine("CLEAR COMMAND BEING PROCESSED\n");

   /* Generate data for an INPUT command with clear as the aid byte */
   cmdData = "&CL";

   /* process the INPUT command */
   processInput(cmdData, nextTran);

   return;
}

/*
 *
 * Routine to handle a WAITATI command.  This command
 * will enable ATI transactions.  Processing of the
 * command file will be suspended, and the program will
 * wait for an ATI transaction to be started by the host.
 *
 */

int processWaitATI()

{
   int          rc;
   char         errmsg[CICS_3270_MAX_EPI_MSG_LEN];
   char         errline[MAX_LINE + 50];
   CICS_EpiATIState_t  ati_on=CICS_EPI_ATI_ON;

   rc = CICS_EpiATIState(TermIndex,&ati_on);

   if (rc == CICS_EPI_NORMAL)
   {
      return 2;
   }
   else
   {
      CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
      sprintf(errline, "%d: CICS_EpiATIState - %s\n",
      rc, errmsg);
      CICS_3270_Write_Log_Num(lineNum, errline);

      return 1;
   } /* endif */
}

/*
 *
 * This routine handles sending data to CICS as a result of an
 * EPI Converse event.
 *
 * Handle inbound 3270 data.  This function is called to obtain data destined
 * for CICS.  The fields in the internal representation with the MDT bit set
 * are obtained and returned.  This function should be called in response
 * to an EPI CONVERSE event.
 *
 * This routine will read lines from the command file until an input
 * command is found (e.g. either BEGINTRAN or INPUT).  Any
 * MODFIELD, DUMPSCR and DUMPFLDS commands found will be executed.
 *
 * If any commands to begin a transaction are found, they are ignored
 * (since a transaction is already running) and an error message is
 * written.
 *
 */

int HandleInbound(char *trnid, char *syst)

{
   char    cmd[MAX_LINE];     /* Next command in command file */
   char    cmdData[MAX_LINE]; /* Command Parameters */
   char    errline[MAX_LINE + 50];

   /*
    * Get the next input command from the command file.
    * Valid commands are INPUT and CLEAR.  Any BEGINTRAN
    * commands are skipped.
    *
    */

   while (1)
   {
      if (nextLine(cmd, cmdData, trnid, syst))
      {
         return 1;      /* end of command file */
      }

      /*
       * Handle the INPUT command
       */

      if (strcmp(cmd,"INPUT") == 0)
      {
         processInput(cmdData, NULL);
         return 0;      /* Indicate successful input */

      } /* endif */

      /*
       * Handle the CLEAR command
       */

      if (strcmp(cmd,"CLEAR") == 0)
      {
         processClear(NULL);
         return 0;      /* Indicate successful input */
      } /* endif */

      /*
       * If we get here, then we have an invalid command.
       * Print an error message and go on to the next line
       * in the command file, looking for an input command.
       */

      sprintf(errline, "Command skipped - not valid for EPI Convserse - %s %s\n",
              cmd, cmdData);
      CICS_3270_Write_Log_Num(lineNum, errline);
   } /* endwhile */
}

/*
 *
 * This routine handles EPI end transaction events.  It
 * searches the command file for the next transaction to
 * start.  This can be either BEGINTRAN or INPUT.
 * If a transaction is sucessfully started, then it will
 * return a zero, otherwise it will return a 1.
 *
 */

int startTrans(char *nextTran, char *syst)

{
   char    trnid[CICS_EPI_TRANSID_MAX + 1]; /* transaction to start */
   char    tranData[MAX_LINE + 3]; /* Transaction data */
   char    cmd[MAX_LINE];          /* Next command in command file */
   char    cmdData[MAX_LINE];      /* Command Parameters */
   int     rc;

   /*
    * Get the next input command from the command file.
    * Valid commands are BEGINTRAN or INPUT.
    * Any other commands are skipped.
    */

   while (inTran == 0)
   {
      if (nextLine(cmd, cmdData, nextTran, syst))
      {
         /* exit if we get to the end of the command file */
         return 1;      /* end of command file */
      }

      /*
       * Handle the BEGINTRAN command
       */

      if (strcmp(cmd,"BEGINTRAN") == 0)
      {
         /* extract the transaction id from cmdData into trnid */
         rc = parseTransid(cmdData, strlen(cmdData), trnid);

         if (rc > 0)
         {
			/* pass the transaction id back to remember the current transid */
			 strcpy(nextTran, trnid);
			 /* create a valid 3270 data stream */
			strcpy(tranData,NEW_TRAN_DATA);  /* aid byte + curspos */
            strcat(tranData, cmdData);       /* append user data */

            /* try and start the transaction */
            epiBeginTran(trnid, tranData, strlen(tranData));
         }
         else
         {
            /* write error to the error log and trace file */
            noTransactionData();
         } /* endif */
      }

      /*
       * Handle the WAITATI command - If successful, this command will
       * return a 2, to indicate that processing of
       * the command file should cease, and wait for
       * an ATI transaction to start.  Note that this
       * command can cause indefinite waits if no ATI
       * command is issued at the host.
       */

      if (strcmp(cmd, "WAITATI") == 0)
      {
         rc = processWaitATI();
         return rc;
      }

      /* Send the data that is in the current buffer to start */
      /* the next transaction.                                */
      if (strcmp(cmd, "INPUT") == 0)
      {
          processInput(cmdData, nextTran);
      }

      if (strcmp(cmd, "CLEAR") == 0)
      {
         processClear(nextTran);
      }
   }   /* End while */

   return 0;
}

void writeHelpInfo(char *cmd)

{
   printf("\n3270 Command Simulator V1.1\n");
   printf("Usage: %s cmdfile <region|* <netname|* <term model> > >\n\n", cmd);
   printf("For trace information, set DEBUG to the name of\n");
   printf("the desired trace file, which will be created or\n");
   printf("overwritten.  To display the trace information on the\n");
   printf("console, set DEBUG=stdout.  Lines beginning with a\n");
   printf("*, ;, ' or / are treated as comments.\n");
   printf("Valid commands are (<> indicates optional):\n\n");
   printf("BEGINTRAN tran <tran data>\n");
   printf("INPUT <aid>\n");
   printf("KEYIN [keystroke data]\n");
   printf("MODFIELD row column data or ");
   printf("MODNUMFLD field-number data\n");
   printf("SETCURSOR row col\n");
   printf("WAITATI\n");
   printf("CLEAR\n");
   printf("DUMPSCR  <filename>\n");
   printf("DUMPFLDS <filename>\n");
   printf("DUMPRAW  filename\n");
   printf("READRAW  filename\n");
   printf("PRINTFLD row column or ");
   printf("PRINTNUMFLD field-number\n");
   printf("TRACE OFF | ON <filename>\n");
   printf("WRITEFIELDS filename\n");
   printf("WRITEPW filename <prefix>\n");
   return;
}

int main(int argc, char *argv[])

{
   int                     rc;
   char                    *regionName = '\0';
   char                    *netName = NULL;
   char                    *termModel = NULL;
   CICS_EpiDetails_t       termDet;
   char                    *filename;
   char                    *data;
   int                     length;
   CICS_EpiEventData_t     Event;
   u_byte_t                dataBuffer[255*255];	/* allocate a large buffer */
   char                    nextTran[CICS_EPI_TRANSID_MAX + 1] = "";
   char                    errmsg[CICS_3270_MAX_EPI_MSG_LEN];
   char                    errline[MAX_LINE + 50];
   CICS_EpiATIState_t      ati_off=CICS_EPI_ATI_HOLD;
   int						epi_term_issued = 0;	/* do not send any more data */

   /**************************************************************/
   /* Check that if a file name was specified                    */
   /**************************************************************/

   if (argc < 2)
   {
      writeHelpInfo(argv[0]);
      return 1;
   }

   filename = argv[1];

   /**************************************************************/
   /* Check for a help request                                   */
   /**************************************************************/


   if ((filename[0] == '?') || (filename[1] == '?'))
   {
      writeHelpInfo(argv[0]);
      return 1;
   } /* endif */

   /**************************************************************/
   /* Check if a CICS region was specified                       */
   /**************************************************************/

   if (argc > 2)
   {
      regionName = argv[2];

      /* check for use of default region */
      if (regionName[0] == '*')
      {
         regionName[0] = '\0';
      } /* endif */
   }

   /**************************************************************/
   /* Check if a CICS network name was specified                 */
   /**************************************************************/

   if (argc > 3)
   {
      netName = argv[3];

      /* check for use of default generated network name */
      if (netName[0] == '*')
      {
         netName = NULL;
      } /* endif */
   }

   /**************************************************************/
   /* Check if a CICS terminal model was specified               */
   /**************************************************************/

   if (argc > 4)
   {
      termModel = argv[4];
   }

   /**************************************************************/
   /* Open the command file.  Exit if the file cannot be         */
   /*  opened.                                                   */
   /**************************************************************/

   if ((cmdFile = fopen(filename, "r")) == NULL)
   {
      printf("Unable to open command file\n");
      return(100);
   }


   /**************************************************************/
   /* Initialize the EPI.                                        */
   /**************************************************************/

   rc = CICS_EpiInitialize(CICS_EPI_VERSION_101);

   if (rc != CICS_EPI_NORMAL)
   {
      /* log the error */
      CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
      sprintf(errline, "%d: CICS_EpiInitialize - %s\n", rc, errmsg);
      CICS_3270_Write_Log_Num(lineNum, errline);

      fclose(cmdFile);
      return 1;
   }

   /**************************************************************/
   /* Install the logical terminal, returning detailed           */
   /* information about the installed terminal.                  */
   /**************************************************************/

   rc = CICS_EpiAddTerminal(
              NULL,
              regionName,
              netName,
              NULL,
              NULL,
              &termDet,
              &TermIndex);

   /**************************************************************/
   /* Check the return code from EpiAddTerminal.                 */
   /**************************************************************/

   if (rc != CICS_EPI_NORMAL)
   {
      /* log the error */
      CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
      sprintf(errline, "%d: CICS_EpiAddTerminal - %s\n", rc,errmsg);
      CICS_3270_Write_Log_Num(lineNum, errline);

      CICS_EpiTerminate();
      fclose(cmdFile);
      return 1;
   }

   /**************************************************************/
   /* If debug is enabled, write out the terminal details        */
   /**************************************************************/

   sprintf(errline,"Terminal Details for Terminal %s Index %d\n",
           termDet.NetName,TermIndex);
   CICS_3270_WriteTraceLine(errline);

   sprintf(errline,"   Lines %d  Columns %d Max Data Size %d",
           termDet.NumLines, termDet.NumColumns, termDet.MaxData);

   if (termDet.Hilight == 1)
   {
      strcat(errline, "  Extended Highlighting");
   } /* endif */

   if (termDet.Color == 1)
   {
      strcat(errline, "  Color");
   } /* endif */

   strcat(errline, "\n");
   CICS_3270_WriteTraceLine(errline);

   /**************************************************************/
   /* Initialize the 3270 data parsing module, specifying a      */
   /*  terminal model to match the installed model.              */
   /**************************************************************/

   switch (termDet.NumLines)
   {
       case 12:
          CICS_3270_Initialize(1);
          break;

       case 24:
          CICS_3270_Initialize(2);
          break;

       case 32:
          CICS_3270_Initialize(3);
          break;

       case 43:
          CICS_3270_Initialize(4);
          break;

       case 27:
          CICS_3270_Initialize(5);
          break;

       default:
          CICS_3270_Initialize(2);
          break;
   } /* endswitch */

   /**************************************************************/
   /* Start executing transactions, setting nextTran to be       */
   /*  NULL.                                                     */
   /**************************************************************/

   memset(nextTran, '\0', sizeof(nextTran));;

   if (startTrans(nextTran, regionName) == 1)
   {
       CICS_EpiTerminate();
       fclose(cmdFile);
       exit(100);   /* No transaction started - exit */
   }

   /**************************************************************/
   /* Enter the main event processing loop, waiting for events   */
   /* from the CICS region to be received.                       */
   /**************************************************************/

   while(1)
   {
   /**************************************************************/
   /* Wait for an event from CICS                                */
   /**************************************************************/

      Event.Data = dataBuffer;
      Event.Size = sizeof(dataBuffer);

      rc = CICS_EpiGetEvent(TermIndex,
                            CICS_EPI_WAIT,
                            &Event);

      if (rc != CICS_EPI_NORMAL && rc != CICS_EPI_ERR_MORE_EVENTS)
      {
         CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
         sprintf(errline, "%d: CICS_EpiGetEvent - %s\n",
                 rc,errmsg);

         CICS_3270_Write_Log_Num(lineNum, errline);

         return 1;
      }

   /**************************************************************/
   /* Handle the expected events.                                */
   /**************************************************************/

      switch(Event.Event)
      {

        /************************************************************************/
        /* The send event indicates that EXEC CICS SEND MAP ... was issued in   */
        /* the transaction.  The data stream received is parsed and modfies the */
        /* internal representation of the terminal.                             */
        /************************************************************************/

         case CICS_EPI_EVENT_SEND:
            CICS_3270_WriteTraceLine("SEND event received\n");

			/* check if we have a request to save the raw input in a file */
			if (strlen(rawFile) > 0)
			{
				/* dump the raw data to the file */
				dumpRawData(Event.Data, Event.Size);

				/* get rid of the current file name */
				rawFile[0] = '\0';
			}

            HandleOutbound(Event.Data, Event.Size);
            break;

        /************************************************************************/
        /* The converse event is expected when an EXEC CICS RECEIVE MAP ... is  */
        /* issued in the transaction.  This may include 3270 data stream not    */
        /* yet transmitted from the CICS region and hence outbound data must be */
        /* handled first.  A CICS RECEIVE BUFFER is recognized and handled.     */
        /************************************************************************/

         case CICS_EPI_EVENT_CONVERSE:
            if ((Event.Size == 1) &&
               (dataBuffer[0] == CICS_3270_CMD_RB))
            {
               /* handle a Read Buffer request */
               CICS_3270_GetBuffer(39, &data, &length);
               rc = CICS_EpiReply(TermIndex,
                                 (u_byte_t *) data,
                                 (u_short_t) length);

               /* check for errors */
               if (rc != CICS_EPI_NORMAL)
               {
                  /* log the error */
                  CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
                  sprintf(errline, "%d: CICS_EpiReply to Read Buffer - %s\n", rc,errmsg);
                  CICS_3270_Write_Log_Num(lineNum, errline);

                  CICS_EpiTerminate();
                  fclose(cmdFile);
                  return 1;
               }
            }
            else
            {
               if (Event.Size > 0)
               {
                  CICS_3270_WriteTraceLine("CONVERSE event received\n");

                  HandleOutbound(Event.Data, Event.Size);
               }

			   /* check if we are terminating */
			   if (epi_term_issued == 0)
			   {
					HandleInbound(nextTran, regionName);
			   }
            }
            break;

        /*****************************************************************/
        /* The end tran event is received when the transaction ends.     */
        /* The command file is then read, looking for the next           */
        /* transaction to execute.  If none is found, the startTrans     */
        /* will return a 1 and the terminal will be deleted.  This       */
        /* will trigger the end of the program.                          */
        /*****************************************************************/

         case CICS_EPI_EVENT_END_TRAN:

            /* no longer in a tran */
            inTran = 0;

            /* remember the next tran id */
            /* if specified on RETURN */
            strcpy(nextTran, Event.TransId);

            sprintf(errline,"END TRAN event received - Next Transid is %s\n",nextTran);
            CICS_3270_WriteTraceLine(errline);

		   /* check if we are terminating */
		   if (epi_term_issued == 0)
		   {
	            /* read the command file and start the next transaction */
		        rc=startTrans(nextTran, regionName);
		   }
		   else
		   {
			   /* we are terminating - don't start another transaction */
			   rc = 0;
		   }

            /* check if we have reached the end of the command file */
            if (rc == 1)
            {
               rc = CICS_EpiDelTerminal(TermIndex);

               if (rc != CICS_EPI_NORMAL)
               {
                  CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
                  sprintf(errline, "%d: CICS_EpiDelTerminal - %s\n",
                          rc, errmsg);
                  CICS_3270_Write_Log_Num(lineNum, errline);
               }
			   else
			   {
				   epi_term_issued = 1;
			   }
            }
            else
            {
               /* Note that rc == 2 if waiting for */
               /* an ATI event                     */
               if (rc == 0)
               {
                  inTran = 1;
               } /* endif */
            }

            break;

        /*******************************************************************/
        /* This event is received when CICS has automatically started      */
        /* a transaction on the terminal.                                  */
        /*******************************************************************/

         case CICS_EPI_EVENT_START_ATI:
            CICS_3270_WriteTraceLine("Start ATI event received\n");
            inTran = 1;   /* indicate transaction active */

            /* now turn ATI off again */
            rc = CICS_EpiATIState(TermIndex,&ati_off);

            if (rc == CICS_EPI_NORMAL)
            {
               ATIEnabled = 0;  /* ATI turned off */
            }
            else
            {
               CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
               sprintf(errline, "%d: CICS_EpiATIState - %s\n",
                       rc, errmsg);
               CICS_3270_Write_Log_Num(lineNum, errline);
            } /* endif */

            break;

        /*******************************************************************/
        /* This event is received when the terminal has been uninstalled.  */
        /*******************************************************************/

         case CICS_EPI_EVENT_END_TERM:
            CICS_3270_WriteTraceLine("END TERMINAL event received\n");

            rc = CICS_EpiTerminate();

            if (rc != CICS_EPI_NORMAL)
            {
               CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
               sprintf(errline, "%d: CICS_EpiTerminate - %s\n",
                       rc, errmsg);
               CICS_3270_Write_Log_Num(lineNum, errline);
            }

            fclose(cmdFile);
            exit(0);

         default:
            CICS_3270_WriteTraceLine("**UNKNOWN** event received\n");

            printf("Unknown Event: %d received\n", Event.Event);
      }
   }
}

/*-------------*/
/* End Of File */
/*-------------*/